from pytest import mark

from graphql import (
    graphql,
    graphql_sync,
    print_schema,
    print_type,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLInputField,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema,
)

from graphql_relay import mutation_with_client_mutation_id

from ..utils import dedent


class Result:

    # noinspection PyPep8Naming
    def __init__(self, result, clientMutationId=None):
        self.clientMutationId = clientMutationId
        self.result = result


def dummy_resolve(_info, inputData=None, clientMutationId=None):
    return Result(inputData or 1, clientMutationId)


async def dummy_resolve_async(_info, inputData=None, clientMutationId=None):
    return Result(inputData or 1, clientMutationId)


def wrap_in_schema(mutation_fields: GraphQLFieldMap) -> GraphQLSchema:
    query_type = GraphQLObjectType("Query", fields={"dummy": GraphQLField(GraphQLInt)})
    mutation_type = GraphQLObjectType("Mutation", fields=mutation_fields)
    return GraphQLSchema(query_type, mutation_type)


def describe_mutation_with_client_mutation_id():
    def requires_an_argument():
        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation", {}, {"result": GraphQLField(GraphQLInt)}, dummy_resolve
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation {
                result
              }
            }
            """
        assert graphql_sync(schema, source) == (
            None,
            [
                {
                    "message": "Field 'someMutation' argument 'input'"
                    " of type 'SomeMutationInput!' is required,"
                    " but it was not provided.",
                    "locations": [(3, 15)],
                }
            ],
        )

    def returns_the_same_client_mutation_id():
        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation", {}, {"result": GraphQLField(GraphQLInt)}, dummy_resolve
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {"someMutation": {"result": 1, "clientMutationId": "abc"}},
            None,
        )

    def supports_thunks_as_input_and_output_fields():
        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            {"inputData": GraphQLInputField(GraphQLInt)},
            {"result": GraphQLField(GraphQLInt)},
            dummy_resolve,
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {inputData: 1234, clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {
                "someMutation": {
                    "result": 1234,
                    "clientMutationId": "abc",
                }
            },
            None,
        )

    @mark.asyncio
    async def supports_async_mutations():
        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            {},
            {"result": GraphQLField(GraphQLInt)},
            dummy_resolve_async,
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert await graphql(schema, source) == (
            {"someMutation": {"result": 1, "clientMutationId": "abc"}},
            None,
        )

    def can_access_root_value():
        some_mutation = mutation_with_client_mutation_id(  # pragma: no cover
            "SomeMutation",
            {},
            {"result": GraphQLField(GraphQLInt)},
            lambda info, clientMutationId=None: Result(
                info.root_value, clientMutationId
            ),
        )

        wrapper_type = GraphQLObjectType("WrapperType", {"someMutation": some_mutation})
        assert print_type(wrapper_type) == dedent(
            """
            type WrapperType {
              someMutation(input: SomeMutationInput!): SomeMutationPayload
            }
            """
        )

    def supports_mutations_returning_null():
        def null_resolve(_info, **_input):
            return None

        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation", {}, {"result": GraphQLField(GraphQLInt)}, null_resolve
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {"someMutation": {"result": None, "clientMutationId": "abc"}},
            None,
        )

    @mark.asyncio
    async def supports_async_mutations_returning_null():
        async def null_resolve(_info, **_input):
            return None

        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            {},
            {"result": GraphQLField(GraphQLInt)},
            null_resolve,
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert await graphql(schema, source) == (
            {"someMutation": {"result": None, "clientMutationId": "abc"}},
            None,
        )

    def supports_mutations_returning_custom_classes():
        class SomeClass:
            @staticmethod
            def get_some_generated_data():
                return 1

            @classmethod
            def mutate(cls, _info, **_input):
                return cls()

            @classmethod
            def resolve(cls, obj, _info):
                assert isinstance(obj, cls)
                return obj.get_some_generated_data()

        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            {},
            {"result": GraphQLField(GraphQLInt, resolve=SomeClass.resolve)},
            SomeClass.mutate,
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {"someMutation": {"result": 1, "clientMutationId": "abc"}},
            None,
        )

    def supports_mutations_returning_mappings():
        def dict_mutate(_info, **_input):
            return {"some_data": 1}

        def dict_resolve(obj, _info):
            return obj["some_data"]

        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            {},
            {"result": GraphQLField(GraphQLInt, resolve=dict_resolve)},
            dict_mutate,
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {"someMutation": {"result": 1, "clientMutationId": "abc"}},
            None,
        )

    @mark.asyncio
    async def supports_async_mutations_returning_mappings():
        async def dict_mutate(_info, **_input):
            return {"some_data": 1}

        async def dict_resolve(obj, _info):
            return obj["some_data"]

        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            {},
            {"result": GraphQLField(GraphQLInt, resolve=dict_resolve)},
            dict_mutate,
        )
        schema = wrap_in_schema({"someMutation": some_mutation})
        source = """
            mutation {
              someMutation(input: {clientMutationId: "abc"}) {
                result
                clientMutationId
              }
            }
            """
        assert await graphql(schema, source) == (
            {"someMutation": {"result": 1, "clientMutationId": "abc"}},
            None,
        )

    def generates_correct_types():
        some_mutation = mutation_with_client_mutation_id(
            "SomeMutation",
            description="Some Mutation Description",
            input_fields={},
            output_fields={"result": GraphQLField(GraphQLInt)},
            mutate_and_get_payload=dummy_resolve,
            deprecation_reason="Just because",
        )

        schema = wrap_in_schema({"someMutation": some_mutation})

        assert print_schema(schema) == dedent(
            '''
            type Query {
              dummy: Int
            }

            type Mutation {
              """Some Mutation Description"""
              someMutation(input: SomeMutationInput!): SomeMutationPayload @deprecated(reason: "Just because")
            }

            type SomeMutationPayload {
              result: Int
              clientMutationId: String
            }

            input SomeMutationInput {
              clientMutationId: String
            }
            '''  # noqa: E501
        )



from typing import List, NamedTuple

from graphql import (
    graphql_sync,
    print_schema,
    GraphQLField,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)

from graphql_relay import (
    backward_connection_args,
    connection_args,
    connection_definitions,
    connection_from_array,
    forward_connection_args,
)

from ..utils import dedent


class User(NamedTuple):
    name: str
    friends: List[int]


all_users = [
    User(name="Dan", friends=[1, 2, 3, 4]),
    User(name="Nick", friends=[0, 2, 3, 4]),
    User(name="Lee", friends=[0, 1, 3, 4]),
    User(name="Joe", friends=[0, 1, 2, 4]),
    User(name="Tim", friends=[0, 1, 2, 3]),
]

friend_connection: GraphQLObjectType
user_connection: GraphQLObjectType

user_type = GraphQLObjectType(
    "User",
    fields=lambda: {
        "name": GraphQLField(GraphQLString),
        "friends": GraphQLField(
            friend_connection,
            args=connection_args,
            resolve=lambda user, _info, **args: connection_from_array(
                user.friends, args
            ),
        ),
        "friendsForward": GraphQLField(
            user_connection,
            args=forward_connection_args,
            resolve=lambda user, _info, **args: connection_from_array(
                user.friends, args
            ),
        ),
        "friendsBackward": GraphQLField(
            user_connection,
            args=backward_connection_args,
            resolve=lambda user, _info, **args: connection_from_array(
                user.friends, args
            ),
        ),
    },
)

friend_connection = connection_definitions(
    GraphQLNonNull(user_type),
    name="Friend",
    resolve_node=lambda edge, _info: all_users[edge.node],
    edge_fields=lambda: {
        "friendshipTime": GraphQLField(
            GraphQLString, resolve=lambda user_, info_: "Yesterday"
        )
    },
    connection_fields=lambda: {
        "totalCount": GraphQLField(
            GraphQLInt, resolve=lambda _user, _info: len(all_users) - 1
        )
    },
).connection_type


user_connection = connection_definitions(
    GraphQLNonNull(user_type), resolve_node=lambda edge, _info: all_users[edge.node]
).connection_type


query_type = GraphQLObjectType(
    "Query",
    fields=lambda: {
        "user": GraphQLField(user_type, resolve=lambda _root, _info: all_users[0])
    },
)

schema = GraphQLSchema(query=query_type)


def describe_connection_definition():
    def includes_connection_and_edge_fields():
        source = """
            {
              user {
                friends(first: 2) {
                  totalCount
                  edges {
                    friendshipTime
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {
                "user": {
                    "friends": {
                        "totalCount": 4,
                        "edges": [
                            {"friendshipTime": "Yesterday", "node": {"name": "Nick"}},
                            {"friendshipTime": "Yesterday", "node": {"name": "Lee"}},
                        ],
                    }
                }
            },
            None,
        )

    def works_with_forward_connection_args():
        source = """
            {
              user {
                friendsForward(first: 2) {
                  edges {
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {
                "user": {
                    "friendsForward": {
                        "edges": [{"node": {"name": "Nick"}}, {"node": {"name": "Lee"}}]
                    }
                }
            },
            None,
        )

    def works_with_backward_connection_args():
        source = """
            {
              user {
                friendsBackward(last: 2) {
                  edges {
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        assert graphql_sync(schema, source) == (
            {
                "user": {
                    "friendsBackward": {
                        "edges": [{"node": {"name": "Joe"}}, {"node": {"name": "Tim"}}]
                    }
                }
            },
            None,
        )

    def generates_correct_types():
        assert print_schema(schema) == dedent(
            '''
            type Query {
              user: User
            }

            type User {
              name: String
              friends(
                """Returns the items in the list that come after the specified cursor."""
                after: String

                """Returns the first n items from the list."""
                first: Int

                """Returns the items in the list that come before the specified cursor."""
                before: String

                """Returns the last n items from the list."""
                last: Int
              ): FriendConnection
              friendsForward(
                """Returns the items in the list that come after the specified cursor."""
                after: String

                """Returns the first n items from the list."""
                first: Int
              ): UserConnection
              friendsBackward(
                """Returns the items in the list that come before the specified cursor."""
                before: String

                """Returns the last n items from the list."""
                last: Int
              ): UserConnection
            }

            """A connection to a list of items."""
            type FriendConnection {
              """Information to aid in pagination."""
              pageInfo: PageInfo!

              """A list of edges."""
              edges: [FriendEdge]
              totalCount: Int
            }

            """Information about pagination in a connection."""
            type PageInfo {
              """When paginating forwards, are there more items?"""
              hasNextPage: Boolean!

              """When paginating backwards, are there more items?"""
              hasPreviousPage: Boolean!

              """When paginating backwards, the cursor to continue."""
              startCursor: String

              """When paginating forwards, the cursor to continue."""
              endCursor: String
            }

            """An edge in a connection."""
            type FriendEdge {
              """The item at the end of the edge"""
              node: User!

              """A cursor for use in pagination"""
              cursor: String!
              friendshipTime: String
            }

            """A connection to a list of items."""
            type UserConnection {
              """Information to aid in pagination."""
              pageInfo: PageInfo!

              """A list of edges."""
              edges: [UserEdge]
            }

            """An edge in a connection."""
            type UserEdge {
              """The item at the end of the edge"""
              node: User!

              """A cursor for use in pagination"""
              cursor: String!
            }
            '''  # noqa: E501
        )



from itertools import chain
from typing import Any, NamedTuple, Optional, Union

from graphql import (
    graphql_sync,
    print_schema,
    GraphQLField,
    GraphQLID,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLString,
)

from graphql_relay import node_definitions

from ..utils import dedent


class User(NamedTuple):
    id: str
    name: str


class Photo(NamedTuple):
    id: str
    width: int


user_data = [User(id="1", name="John Doe"), User(id="2", name="Jane Smith")]

photo_data = [Photo(id="3", width=300), Photo(id="4", width=400)]


def get_node(id_: str, info: GraphQLResolveInfo) -> Optional[Union[User, Photo]]:
    assert info.schema is schema
    return next(
        filter(
            lambda obj: obj.id == id_,  # type: ignore
            chain(user_data, photo_data),
        ),
        None,
    )


def get_node_type(
    obj: Union[User, Photo], info: GraphQLResolveInfo, _type: Any
) -> Optional[str]:
    assert info.schema is schema
    if obj in user_data:
        return user_type.name
    if obj in photo_data:
        return photo_type.name
    return None  # pragma: no cover


node_interface, node_field, nodes_field = node_definitions(get_node, get_node_type)


user_type = GraphQLObjectType(
    "User",
    lambda: {
        "id": GraphQLField(GraphQLNonNull(GraphQLID)),
        "name": GraphQLField(GraphQLString),
    },
    interfaces=[node_interface],
)

photo_type = GraphQLObjectType(
    "Photo",
    lambda: {
        "id": GraphQLField(GraphQLNonNull(GraphQLID)),
        "width": GraphQLField(GraphQLInt),
    },
    interfaces=[node_interface],
)

query_type = GraphQLObjectType(
    "Query", lambda: {"node": node_field, "nodes": nodes_field}
)

schema = GraphQLSchema(query=query_type, types=[node_interface, user_type, photo_type])


def describe_node_interface_and_fields():
    def describe_ability_to_refetch():
        def gets_the_correct_id_for_users():
            source = """
              {
                node(id: "1") {
                  id
                }
              }
            """
            assert graphql_sync(schema, source) == ({"node": {"id": "1"}}, None)

        def gets_the_correct_name_for_users():
            source = """
              {
                node(id: "1") {
                  id
                  ... on User {
                    name
                  }
                }
              }
            """
            assert graphql_sync(schema, source) == (
                {"node": {"id": "1", "name": "John Doe"}},
                None,
            )

        def gets_the_correct_width_for_photos():
            source = """
              {
                node(id: "4") {
                  id
                  ... on Photo {
                    width
                  }
                }
              }
            """
            assert graphql_sync(schema, source) == (
                {"node": {"id": "4", "width": 400}},
                None,
            )

        def gets_the_correct_typename_for_users():
            source = """
              {
                node(id: "1") {
                  id
                  __typename
                }
              }
            """
            assert graphql_sync(schema, source) == (
                {"node": {"id": "1", "__typename": "User"}},
                None,
            )

        def gets_the_correct_typename_for_photos():
            source = """
              {
                node(id: "4") {
                  id
                  __typename
                }
              }
            """
            assert graphql_sync(schema, source) == (
                {"node": {"id": "4", "__typename": "Photo"}},
                None,
            )

        def ignores_photo_fragments_on_user():
            source = """
              {
                node(id: "1") {
                  id
                  ... on Photo {
                    width
                  }
                }
              }
            """
            assert graphql_sync(schema, source) == ({"node": {"id": "1"}}, None)

        def returns_null_for_bad_ids():
            source = """
              {
                node(id: "5") {
                  id
                }
              }
            """
            assert graphql_sync(schema, source) == ({"node": None}, None)

        def returns_nulls_for_bad_ids():
            source = """
              {
                nodes(ids: ["3", "5"]) {
                  id
                }
              }
            """
            assert graphql_sync(schema, source) == (
                {"nodes": [{"id": "3"}, None]},
                None,
            )

    def generates_correct_types():
        assert print_schema(schema) == dedent(
            '''
            """An object with an ID"""
            interface Node {
              """The id of the object."""
              id: ID!
            }

            type User implements Node {
              id: ID!
              name: String
            }

            type Photo implements Node {
              id: ID!
              width: Int
            }

            type Query {
              """Fetches an object given its ID"""
              node(
                """The ID of an object"""
                id: ID!
              ): Node

              """Fetches objects given their IDs"""
              nodes(
                """The IDs of objects"""
                ids: [ID!]!
              ): [Node]!
            }
            '''
        )



import warnings

# noinspection PyDeprecation
from .array_connection import (
    connection_from_array,
    connection_from_array_slice,
    cursor_for_object_in_connection,
    cursor_to_offset,
    get_offset_with_default,
    offset_to_cursor,
    SizedSliceable,
)

warnings.warn(
    "The 'arrayconnection' module is deprecated. "
    "Functions should be imported from the top-level package instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = [
    "connection_from_array",
    "connection_from_array_slice",
    "cursor_for_object_in_connection",
    "cursor_to_offset",
    "get_offset_with_default",
    "offset_to_cursor",
    "SizedSliceable",
]



from graphql import graphql_sync

from .star_wars_schema import star_wars_schema as schema


def describe_star_wars_object_identification():
    def fetches_the_id_and_name_of_the_rebels():
        source = """
            {
              rebels {
                id
                name
              }
            }
            """
        expected = {
            "rebels": {"id": "RmFjdGlvbjox", "name": "Alliance to Restore the Republic"}
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_the_rebels_by_global_id():
        source = """
            {
              node(id: "RmFjdGlvbjox") {
                id
                ... on Faction {
                  name
                }
              }
            }
            """
        expected = {
            "node": {"id": "RmFjdGlvbjox", "name": "Alliance to Restore the Republic"}
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_the_id_and_name_of_the_empire():
        source = """
            {
              empire {
                id
                name
              }
            }
            """
        expected = {"empire": {"id": "RmFjdGlvbjoy", "name": "Galactic Empire"}}
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_the_empire_by_global_id():
        source = """
            {
              node(id: "RmFjdGlvbjoy") {
                id
                ... on Faction {
                  name
                }
              }
            }
            """
        expected = {"node": {"id": "RmFjdGlvbjoy", "name": "Galactic Empire"}}
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_the_x_wing_by_global_id():
        source = """
            {
              node(id: "U2hpcDox") {
                id
                ... on Ship {
                  name
                }
              }
            }
            """
        expected = {"node": {"id": "U2hpcDox", "name": "X-Wing"}}
        result = graphql_sync(schema, source)
        assert result == (expected, None)



from typing import NamedTuple

from graphql import (
    graphql_sync,
    print_schema,
    GraphQLField,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLString,
)

from graphql_relay import plural_identifying_root_field

from ..utils import dedent

user_type = GraphQLObjectType(
    "User",
    fields=lambda: {
        "username": GraphQLField(GraphQLString),
        "url": GraphQLField(GraphQLString),
    },
)


class User(NamedTuple):
    username: str
    url: str


def resolve_single_input(info: GraphQLResolveInfo, username: str) -> User:
    assert info.schema is schema
    lang = info.context.lang
    url = f"www.facebook.com/{username}?lang={lang}"
    return User(username=username, url=url)


query_type = GraphQLObjectType(
    "Query",
    lambda: {
        "usernames": plural_identifying_root_field(
            "usernames",
            description="Map from a username to the user",
            input_type=GraphQLString,
            output_type=user_type,
            resolve_single_input=resolve_single_input,
        )
    },
)

schema = GraphQLSchema(query=query_type)


class Context(NamedTuple):
    lang: str


def describe_plural_identifying_root_field():
    def allows_fetching():
        source = """
        {
          usernames(usernames:["dschafer", "leebyron", "schrockn"]) {
            username
            url
          }
        }
        """
        context_value = Context(lang="en")
        assert graphql_sync(schema, source, context_value=context_value) == (
            {
                "usernames": [
                    {
                        "username": "dschafer",
                        "url": "www.facebook.com/dschafer?lang=en",
                    },
                    {
                        "username": "leebyron",
                        "url": "www.facebook.com/leebyron?lang=en",
                    },
                    {
                        "username": "schrockn",
                        "url": "www.facebook.com/schrockn?lang=en",
                    },
                ]
            },
            None,
        )

    def generates_correct_types():
        assert print_schema(schema) == dedent(
            '''
            type Query {
              """Map from a username to the user"""
              usernames(usernames: [String!]!): [User]
            }

            type User {
              username: String
              url: String
            }
            '''
        )



from graphql import graphql_sync

from .star_wars_schema import star_wars_schema as schema


def describe_star_wars_connections():
    def fetches_the_first_ship_of_the_rebels():
        source = """
            {
              rebels {
                name,
                ships(first: 1) {
                  edges {
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        expected = {
            "rebels": {
                "name": "Alliance to Restore the Republic",
                "ships": {"edges": [{"node": {"name": "X-Wing"}}]},
            }
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_the_first_two_ships_of_the_rebels_with_a_cursor():
        source = """
            {
              rebels {
                name,
                ships(first: 2) {
                  edges {
                    cursor,
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        expected = {
            "rebels": {
                "name": "Alliance to Restore the Republic",
                "ships": {
                    "edges": [
                        {
                            "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                            "node": {"name": "X-Wing"},
                        },
                        {
                            "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                            "node": {"name": "Y-Wing"},
                        },
                    ]
                },
            }
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_the_next_three_ships_of_the_rebels_with_a_cursor():
        source = """
            {
              rebels {
                name,
                ships(first: 3 after: "YXJyYXljb25uZWN0aW9uOjE=") {
                  edges {
                    cursor,
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        expected = {
            "rebels": {
                "name": "Alliance to Restore the Republic",
                "ships": {
                    "edges": [
                        {
                            "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                            "node": {"name": "A-Wing"},
                        },
                        {
                            "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                            "node": {"name": "Millennium Falcon"},
                        },
                        {
                            "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                            "node": {"name": "Home One"},
                        },
                    ]
                },
            }
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def fetches_no_ships_of_the_rebels_at_the_end_of_connection():
        source = """
            {
              rebels {
                name,
                ships(first: 3 after: "YXJyYXljb25uZWN0aW9uOjQ=") {
                  edges {
                    cursor,
                    node {
                      name
                    }
                  }
                }
              }
            }
            """
        expected = {
            "rebels": {
                "name": "Alliance to Restore the Republic",
                "ships": {"edges": []},
            }
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)

    def identifies_the_end_of_the_list():
        source = """
            {
              rebels {
                name,
                originalShips: ships(first: 2) {
                  edges {
                    node {
                      name
                    }
                  }
                  pageInfo {
                    hasNextPage
                  }
                }
                moreShips: ships(first: 3 after: "YXJyYXljb25uZWN0aW9uOjE=") {
                  edges {
                    node {
                      name
                    }
                  }
                  pageInfo {
                    hasNextPage
                  }
                }
              }
            }
            """
        expected = {
            "rebels": {
                "name": "Alliance to Restore the Republic",
                "originalShips": {
                    "edges": [
                        {
                            "node": {"name": "X-Wing"},
                        },
                        {
                            "node": {"name": "Y-Wing"},
                        },
                    ],
                    "pageInfo": {"hasNextPage": True},
                },
                "moreShips": {
                    "edges": [
                        {
                            "node": {"name": "A-Wing"},
                        },
                        {
                            "node": {"name": "Millennium Falcon"},
                        },
                        {
                            "node": {"name": "Home One"},
                        },
                    ],
                    "pageInfo": {"hasNextPage": False},
                },
            },
        }
        result = graphql_sync(schema, source)
        assert result == (expected, None)
